class CanaryMiddleware:
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        environ['repoze.debug.canary'] = Canary()
        return self.app(environ, start_response)

class Canary(object):
    pass

def make_middleware(app, global_conf):
    """ Paste filter-app converter """
    return CanaryMiddleware(app)



from cStringIO import StringIO
from sys import _current_frames
import thread
import time
import traceback

import webob

def dump_threads():
    """Dump running threads

    Returns a string with the tracebacks.
    """

    frames = _current_frames()
    this_thread_id = thread.get_ident()
    now = time.strftime("%Y-%m-%d %H:%M:%S")
    res = ["Threads traceback dump at %s\n" % now]
    for thread_id, frame in frames.iteritems():
        if thread_id == this_thread_id:
            continue

        # Find request in frame
        reqinfo = ''
        f = frame
        while f is not None:
            co = f.f_code

           #reqinfo += '\n\t ' + co.co_name + '\t ' + co.co_filename

            if (co.co_name == 'publish' and
               co.co_filename.endswith('publisher/publish.py')):
                request = f.f_locals.get('request')
                if request is not None:
                    reqinfo += (request.get('REQUEST_METHOD', '') + ' ' +
                               request.get('PATH_INFO', ''))
                    qs = request.get('QUERY_STRING')
                    if qs:
                        reqinfo += '?'+qs
                break
            f = f.f_back
        if reqinfo:
            reqinfo = " (%s)" % reqinfo

        output = StringIO()
        traceback.print_stack(frame, file=output)
        res.append("Thread %s%s:\n%s" %
            (thread_id, reqinfo, output.getvalue()))

    frames = None
    res.append("End of dump")
    return '\n'.join(res)

class MonitoringMiddleware(object):
    """The monitoring middleware intercepts requests for the path
    '/debug_threads' and returns a plain-text thread dump."""
    
    def __init__(self, app):
        self.app = app
        
    def __call__(self, environ, start_response):
        request = webob.Request(environ)

        if request.path == '/debug_threads':
            response = webob.Response(request=request)
            response.content_type = 'text/plain'
            t = dump_threads()
            if isinstance(t, unicode):
                response.unicode_body = t
            else:
                response.body = t
        else:
            response = request.get_response(self.app, catch_exc_info=True)
            
        return response(environ, start_response)

def make_middleware(app, global_conf):
    return MonitoringMiddleware(app)



import unittest

class TestCanaryMiddleware(unittest.TestCase):
    def _getTargetClass(self):
        from repoze.debug.canary import CanaryMiddleware
        return CanaryMiddleware

    def _makeOne(self, app):
        klass = self._getTargetClass()
        return klass(app)

    def test_call(self):
        def app(environ, start_response):
            environ['app_saw'] = True
            return True
        mw = self._makeOne(app)
        environ = {}
        result = mw(environ, None)
        self.assertEqual(result, True)
        from repoze.debug.canary import Canary
        self.assertTrue(isinstance(environ['repoze.debug.canary'], Canary))
        self.assertEqual(environ['app_saw'], True)
        
class TestMakeCanaryMiddleware(unittest.TestCase):
    def _getFUT(self):
        from repoze.debug.canary import make_middleware
        return make_middleware

    def test_maker(self):
        f = self._getFUT()
        mw = f(None, None)
        self.assertEqual(mw.app, None)
        
        


        
        



import unittest

class TestPDBPM(unittest.TestCase):
    def _getFUT(self):
        from repoze.debug.pdbpm import PostMortemDebug
        return PostMortemDebug

    def _makeOne(self, app, *ignore_exc):
        f = self._getFUT()
        return f(app, *ignore_exc)

    def test_post_mortem_withexc(self):
        app = DummyApplication(KeyError)
        mw = self._makeOne(app)
        fake_pdb = FakePDB()
        try:
            import repoze.debug.pdbpm
            old_pdb = repoze.debug.pdbpm.pdb
            repoze.debug.pdbpm.pdb = fake_pdb
            environ = {}
            self.assertRaises(KeyError, mw, environ, None)
        finally:
            repoze.debug.pdb = old_pdb
        self.assertEqual(fake_pdb.called, True)

    def test_post_mortem_noexc(self):
        app = DummyApplication()
        mw = self._makeOne(app)
        fake_pdb = FakePDB()
        try:
            import repoze.debug.pdbpm
            old_pdb = repoze.debug.pdbpm.pdb
            repoze.debug.pdbpm.pdb = fake_pdb
            environ = {}
            result = mw(environ, None)
        finally:
            repoze.debug.pdbpm.pdb = old_pdb
        self.assertEqual(fake_pdb.called, False)
        self.assertEqual(result, ['hello world'])

    def test_post_mortem_ignore_exc(self):
        app = DummyApplication(KeyError)
        mw = self._makeOne(app, KeyError)
        fake_pdb = FakePDB()
        try:
            import repoze.debug.pdbpm
            old_pdb = repoze.debug.pdbpm.pdb
            repoze.debug.pdbpm.pdb = fake_pdb
            environ = {}
            self.assertRaises(KeyError, mw, environ, None)
        finally:
            repoze.debug.pdb = old_pdb
        self.assertEqual(fake_pdb.called, False)

class TestMakeMiddleware(unittest.TestCase):
    def _callFUT(self, app, global_conf=None, ignore_http_exceptions=True):
        from repoze.debug.pdbpm import make_middleware
        return make_middleware(app, global_conf, ignore_http_exceptions)

    def test_paste_constructor_ignore_http_exceptions(self):
        app = DummyApplication() 
        mw = self._callFUT(app, None)
        self.assertEqual(mw.__name__, 'middleware')

    def test_paste_constructor_ignore_http_exceptions_False(self):
        app = DummyApplication() 
        mw = self._callFUT(app, None, False)
        self.assertEqual(mw.__name__, 'middleware')
    

class FakePDB:
    called = False
    def post_mortem(self, *args):
        self.called = True

class DummyApplication:
    def __init__(self, exc=None):
        self.exc = exc

    def __call__(self, environ, start_response):
        if self.exc:
            raise self.exc
        self.environ = environ
        self.start_response = start_response
        return ['hello world']

class DummyException(Exception):
    pass



import unittest

class TestResponseLoggingMiddleware(unittest.TestCase):
    def _getTargetClass(self):
        from repoze.debug.responselogger import ResponseLoggingMiddleware
        return ResponseLoggingMiddleware

    def _makeOne(self, *arg, **kw):
        klass = self._getTargetClass()
        return klass(*arg, **kw)

    def _makeEnviron(self):
        import io
        environ = {
            'SERVER_NAME':'localhost',
            'SERVER_PORT':'80',
            'wsgi.version':(1,0),
            'wsgi.multiprocess':False,
            'wsgi.multithread':True,
            'wsgi.run_once':False,
            'wsgi.url_scheme':'http',
            'wsgi.input':io.BytesIO(b'hello world'),
            }
        return environ

    def test_call(self):
        body = ['thebody']
        app = DummyApp(body, '200 OK', [('HeaderKey', 'headervalue')])
        vlogger = FakeLogger()
        tlogger = FakeLogger()
        mw = self._makeOne(app, 0, 10, vlogger, tlogger)
        environ = self._makeEnviron()
        start_response = FakeStartResponse()
        app_iter = mw(environ, start_response)
        self.assertEqual(''.join(list(app_iter)), 'thebody')
        self.assertEqual(len(vlogger.logged), 2)
        self.assertEqual(start_response.status, '200 OK')
        self.assertEqual(len(start_response.headers), 1)
        self.assertEqual(start_response.headers[0], ('HeaderKey','headervalue'))
        self.assertEqual(start_response.exc_info, None)
        self.assertEqual(app.called, True)

    def test_call_gui_url(self):
        body = ['thebody']
        app = DummyApp(body, '200 OK', [('HeaderKey', 'headervalue')])
        vlogger = FakeLogger()
        tlogger = FakeLogger()
        mw = self._makeOne(app, 0, 10, vlogger, tlogger)
        environ = self._makeEnviron()
        environ['PATH_INFO'] = '/__repoze.debug/feed.xml'
        environ['REQUEST_METHOD'] = 'GET'
        start_response = FakeStartResponse()
        app_iter = mw(environ, start_response)

    def test_call_nologgers(self):
        body = ['thebody']
        app = DummyApp(body, '200 OK', [('HeaderKey', 'headervalue')])
        mw = self._makeOne(app, 0, 10, None, None)
        environ = self._makeEnviron()
        start_response = FakeStartResponse()
        app_iter = mw(environ, start_response)
        self.assertEqual(''.join(list(app_iter)), 'thebody')
        self.assertEqual(start_response.status, '200 OK')
        self.assertEqual(len(start_response.headers), 1)
        self.assertEqual(start_response.headers[0], ('HeaderKey','headervalue'))
        self.assertEqual(start_response.exc_info, None)
        self.assertEqual(app.called, True)

    def test_call_overmaxbodylen(self):
        body = ['thebody']
        app = DummyApp(body, '200 OK', [('HeaderKey', 'headervalue')])
        vlogger = FakeLogger()
        tlogger = FakeLogger()
        mw = self._makeOne(app, 1, 10, vlogger, tlogger)
        environ = self._makeEnviron()
        start_response = FakeStartResponse()
        app_iter = mw(environ, start_response)
        self.assertEqual(''.join(app_iter), 'thebody')
        self.assertEqual(len(vlogger.logged), 2)
        self.assertTrue('(truncated at 1 bytes)' in vlogger.logged[1])

    def test_call_overkeep(self):
        body = ['thebody']
        app = DummyApp(body, '200 OK', [('HeaderKey', 'headervalue')])
        vlogger = FakeLogger()
        tlogger = FakeLogger()
        mw = self._makeOne(app, 1, 1, vlogger, tlogger)
        mw.entries = ['a']
        environ = self._makeEnviron()
        start_response = FakeStartResponse()
        app_iter = mw(environ, start_response)
        self.assertEqual(len(mw.entries), 1)

    def test_call_keep_zero_doesnt_append_entry(self):
        body = ['thebody']
        app = DummyApp(body, '200 OK', [('HeaderKey', 'headervalue')])
        vlogger = FakeLogger()
        tlogger = FakeLogger()
        mw = self._makeOne(app, 1, 0, vlogger, tlogger)
        mw.entries = []
        environ = self._makeEnviron()
        start_response = FakeStartResponse()
        app_iter = mw(environ, start_response)
        self.assertEqual(len(mw.entries), 0)

    def test_call_start_response_not_called(self):
        body = ['thebody']
        app = DummyBrokenApp(body, '200 OK', [('HeaderKey', 'headervalue')])
        vlogger = FakeLogger()
        tlogger = FakeLogger()
        mw = self._makeOne(app, 1, 1, vlogger, tlogger)
        environ = self._makeEnviron()
        start_response = FakeStartResponse()
        app_iter = mw(environ, start_response)
        self.assertEqual(mw.entries[-1]['response']['status'],
                         '500 Start Response Not Called')

    def test_call_app_iter_close(self):
        class Iterable(object):
            def __init__(self, data):
                self.data = data
            def close(self):
                self.closed = True
            def __iter__(self):
                return self
            def next(self):
                if not self.data:
                    raise StopIteration
                return self.data.pop(0)
            __next__ = next
        iterable = Iterable(['1', '2'])
        app = DummyBrokenApp(iterable, '200 OK', [('HeaderKey', 'headervalue')])
        vlogger = FakeLogger()
        tlogger = FakeLogger()
        mw = self._makeOne(app, 1, 1, vlogger, tlogger)
        environ = self._makeEnviron()
        start_response = FakeStartResponse()
        app_iter = list(mw(environ, start_response))
        self.assertEqual(iterable.closed, True)

    def test_call_contentlengthwrong(self):
        body = ['thebody']
        app = DummyApp(body, '200 OK', [('Content-Length', '1')])
        vlogger = FakeLogger()
        tlogger = FakeLogger()
        mw = self._makeOne(app, 1, 10, vlogger, tlogger)
        environ = self._makeEnviron()
        start_response = FakeStartResponse()
        app_iter = mw(environ, start_response)
        self.assertEqual(''.join(app_iter), 'thebody')
        self.assertEqual(len(vlogger.logged), 2)
        self.assertTrue('WARNING-1' in vlogger.logged[1])

    def test_call_sourceurl_in_response(self):
        body = ['thebody']
        app = DummyApp(body, '200 OK', [('Content-Length', '1')])
        vlogger = FakeLogger()
        tlogger = FakeLogger()
        mw = self._makeOne(app, 1, 10, vlogger, tlogger)
        environ = self._makeEnviron()
        start_response = FakeStartResponse()
        app_iter = mw(environ, start_response)
        self.assertEqual(''.join(app_iter), 'thebody')
        self.assertEqual(len(vlogger.logged), 2)
        self.assertTrue('URL: GET http://localhost' in vlogger.logged[1])

    def test_entry_created(self):
        body = ['thebody']
        app = DummyApp(body, '200 OK', [('Content-Length', '1')])
        vlogger = FakeLogger()
        tlogger = FakeLogger()
        mw = self._makeOne(app, 1, 10, vlogger, tlogger)
        start_response = FakeStartResponse()
        environ = self._makeEnviron()
        app_iter = mw(environ, start_response)
        self.assertEqual(''.join(app_iter), 'thebody')
        self.assertEqual(len(mw.entries), 1)
        entry = mw.entries[0]
        self.assertEqual(entry['response']['status'], '200 OK')
        self.assertTrue(isinstance(entry['request']['begin'], float))
        self.assertTrue(isinstance(entry['response']['begin'], float))
        self.assertTrue(isinstance(entry['response']['end'], float))
        self.assertEqual(entry['response']['headers'],
                         [('Content-Length', '1')])
        self.assertEqual(entry['request']['url'], 'http://localhost')
        self.assertEqual(entry['request']['body'], b'hello world')
        self.assertEqual(entry['response']['content-length'], 1)
        self.assertEqual(len(entry['request']['cgi_variables']), 2)
        self.assertEqual(len(entry['request']['wsgi_variables']), 2)
        self.assertTrue(isinstance(entry['id'], int))

    def test_trace_logging(self):
        import time
        body = ['thebody']
        app = DummyApp(body, '200 OK', [('Content-Length', '7')])
        vlogger = FakeLogger()
        tlogger = FakeLogger()
        mw = self._makeOne(app, 1, 10, vlogger, tlogger)
        start_response = FakeStartResponse()
        environ = self._makeEnviron()
        mw.pid = 0
        mw._now = now = time.time()
        app_iter = mw(environ, start_response)
        self.assertEqual(''.join(app_iter), 'thebody')
        self.assertEqual(len(tlogger.logged), 4)
        logged = tlogger.logged
        entry = mw.entries[0]
        rid = entry['id']
        begin = entry['request']['begin']
        self.assertEqual(begin, now)
        
        result = logged[0].split(' ', 4)
        self.assertEqual(result[0], 'U')
        self.assertEqual(result[1], str(mw.pid))
        self.assertEqual(result[2], str(rid))
        self.assertEqual(result[3], str(begin))

        result = logged[1].split(' ', 4)
        self.assertEqual(result[0], 'B')
        self.assertEqual(result[1], str(mw.pid))
        self.assertEqual(result[2], str(rid))
        self.assertEqual(result[3], str(begin))
        self.assertEqual(result[4], 'GET http://localhost')
                         
        result = logged[2].split(' ', 4)
        self.assertEqual(result[0], 'A')
        self.assertEqual(result[1], str(mw.pid))
        self.assertEqual(result[2], str(rid))
        self.assertEqual(result[3], str(begin))
        self.assertEqual(result[4], '200 7')

        result = logged[3].split(' ', 4)
        self.assertEqual(result[0], 'E')
        self.assertEqual(result[1], str(mw.pid))
        self.assertEqual(result[2], str(rid))
        self.assertEqual(result[3], str(begin))
        self.assertEqual(result[4], '7')

class TestMakeResponseLoggingMiddleware(unittest.TestCase):
    def _getFUT(self):
        from repoze.debug.responselogger import make_middleware
        return make_middleware

    def test_make_middleware_defaults(self):
        f = self._getFUT()
        app = DummyApp(None, None, None)
        global_conf = {}
        import tempfile
        fn = tempfile.mktemp()
        mw = f(app, global_conf)
        self.assertEqual(mw.verbose_logger, None)
        self.assertEqual(mw.max_bodylen, 3072)
        self.assertEqual(mw.keep, 100)

    def test_make_middleware_nondefaults(self):
        f = self._getFUT()
        app = DummyApp(None, None, None)
        global_conf = {}
        import tempfile
        vfn = tempfile.mktemp()
        tfn = tempfile.mktemp()
        mw = f(app, global_conf, vfn, tfn, '0', '0', '0', '0')
        self.assertEqual(len(mw.verbose_logger.handlers), 1)
        self.assertEqual(len(mw.trace_logger.handlers), 1)
        self.assertEqual(mw.max_bodylen, 0)
        self.assertEqual(mw.keep, 0)
        mw.verbose_logger.handlers[0].close()
        mw.trace_logger.handlers[0].close()

class FakeStartResponse:
    def __call__(self, status, headers, exc_info=None):
        self.status = status
        self.headers = headers
        self.exc_info = exc_info

class FakeLogger:
    def __init__(self):
        self.logged = []
        
    def info(self, msg):
        self.logged.append(msg)

class DummyApp:
    def __init__(self, body, status, headers):
        self.body = body
        self.status = status
        self.headers = headers
        
    def __call__(self, environ, start_response):
        start_response(self.status, self.headers)
        self.called = True
        return self.body
    
class DummyBrokenApp(DummyApp):
    def __call__(self,environ, start_response):
        self.called = True
        return self.body



import unittest

class Test_is_gui_url(unittest.TestCase):

    def _callFUT(self, environ):
        from repoze.debug.ui import is_gui_url
        return is_gui_url(environ)

    def test_no_PATH_INFO(self):
        self.assertFalse(self._callFUT({}))

    def test_w_PATH_INFO_not_gui_flag(self):
        self.assertFalse(self._callFUT({'PATH_INFO': 'something'}))

    def test_w_PATH_INFO_w_gui_flag(self):
        self.assertTrue(self._callFUT({'PATH_INFO': '__repoze.debug/foo'}))

class Test_get_mimetype(unittest.TestCase):

    def _callFUT(self, filename):
        from repoze.debug.ui import get_mimetype
        return get_mimetype(filename)

    def test_known_type(self):
        self.assertEqual(self._callFUT('foo.txt'), 'text/plain')

    def test_unknown_type(self):
        self.assertEqual(self._callFUT('foo.weird'),
                         'application/octet-stream')

    def test_unknown_type_XUL(self):
        self.assertEqual(self._callFUT('foo.bar.xul'),
                         'application/vnd.mozilla.xul+xml')

class DebugGuiTests(unittest.TestCase):

    def _getTargetClass(self):
        from repoze.debug.ui import DebugGui
        return DebugGui

    def _makeOne(self, middleware):
        return self._getTargetClass()(middleware)

    def _makeEnviron(self, **kw):
        environ = {'PATH_INFO': '/',
                   'REQUEST_METHOD': 'GET',
                  }
        environ.update(kw)
        return environ

    def _make_start_response(self):
        _started = []
        def _start_response(*args):
            _started.append(args)
        return _started, _start_response

    def test___call___static_missing(self):
        environ = self._makeEnviron(PATH_INFO='/static/nonesuch.html')
        _started, _start_response = self._make_start_response()
        gui = self._makeOne(None)
        self.assertRaises(ValueError, gui, environ, _start_response)

    def test___call___static_not_missing(self):
        environ = self._makeEnviron(PATH_INFO='/static/debugui.html')
        _started, _start_response = self._make_start_response()
        gui = self._makeOne(None)
        app_iter =  gui(environ, _start_response)
        self.assertEqual(len(_started), 1)
        self.assertEqual(_started[0][0], '200 OK')
        headers = _started[0][1]
        self.assertTrue(('Content-Type', 'text/html; charset=UTF-8')
                            in headers)
        self.assertTrue(b'<html' in b''.join(list(app_iter)))
        # XXX need more assertions?  Damn trying to test rendered output!

    def test___call___unknown(self):
        environ = self._makeEnviron(PATH_INFO='/nonesuch.html')
        _started, _start_response = self._make_start_response()
        gui = self._makeOne(None)
        self.assertRaises(ValueError, gui, environ, _start_response)

    def test___call___w_feed_xml(self):
        environ = self._makeEnviron(PATH_INFO='/__repoze.debug/feed.xml')
        _started, _start_response = self._make_start_response()
        mw = DummyModel(entries=[], pid=1234)
        gui = self._makeOne(mw)
        app_iter =  gui(environ, _start_response)
        self.assertEqual(len(_started), 1)
        self.assertEqual(_started[0][0], '200 OK')
        headers = _started[0][1]
        self.assertTrue(('Content-Type', 'application/atom+xml; charset=UTF-8')
                            in headers)
        self.assertTrue(b'<atom:feed' in b''.join(list(app_iter)))
        # XXX need more assertions?  Damn trying to test rendered output!

    def test_getStatic_miss(self):
        gui = self._makeOne(None)
        self.assertRaises(ValueError, gui.getStatic, 'nonesuch.html')

    def test_getStatic_hit(self):
        from webob import Response
        gui = self._makeOne(None)
        response = gui.getStatic('debugui.html')
        self.assertTrue(isinstance(response, Response))
        self.assertEqual(response.content_type, 'text/html')
        self.assertEqual(response.charset, 'UTF-8')
        # XXX need more assertions?  Damn trying to test rendered output!

    def test_getFeed_empty(self):
        from webob import Response
        mw = DummyModel(entries=[], pid=1234)
        gui = self._makeOne(mw)
        response = gui.getFeed()
        self.assertTrue(isinstance(response, Response))
        self.assertEqual(response.content_type, 'application/atom+xml')
        # XXX need more assertions?  Damn trying to test rendered output!

    def test_getFeed_non_empty_no_response(self):
        from webob import Response
        entries = [
            {'id': 'aaaa',
             'request': {
                'begin': 1234,
                'method': 'GET',
                'url': '/foo',
                'cgi_variables': [('cgi_a', 'CGI_A')],
                'wsgi_variables': [('wsgi_a', 'WSGI_A')],
                'body': '',
                },
             #'response': {},
            },
        ]
        mw = DummyModel(entries=entries, pid=1234)
        gui = self._makeOne(mw)
        response = gui.getFeed()
        self.assertTrue(isinstance(response, Response))
        self.assertEqual(response.content_type, 'application/atom+xml')
        # XXX need more assertions?  Damn trying to test rendered output!

    def test_getFeed_non_empty_no_response_long_url(self):
        from webob import Response
        entries = [
            {'id': 'aaaa',
             'request': {
                'begin': 1234,
                'method': 'GET',
                'url': '/foo/%s' % '/'.join(['bbb' * 40]),
                'cgi_variables': [('cgi_a', 'CGI_A')],
                'wsgi_variables': [('wsgi_a', 'WSGI_A')],
                'body': '',
                },
             #'response': {},
            },
        ]
        mw = DummyModel(entries=entries, pid=1234)
        gui = self._makeOne(mw)
        response = gui.getFeed()
        self.assertTrue(isinstance(response, Response))
        self.assertEqual(response.content_type, 'application/atom+xml')
        # XXX need more assertions?  Damn trying to test rendered output!

    def test_getFeed_non_empty_w_response(self):
        from webob import Response
        entries = [
            {'id': 'aaaa',
             'request': {
                'begin': 1234,
                'method': 'GET',
                'url': '/foo',
                'cgi_variables': [('cgi_a', 'CGI_A')],
                'wsgi_variables': [('wsgi_a', 'WSGI_A')],
                'body': '',
                },
             'response': {
                'begin': 1235,
                'end': 1236,
                'status': '200 OK',
                'content-length': 57,
                'headers': [('header_a', 'HEADER_A')],
                'body': 'X' * 57,
                },
            },
        ]
        mw = DummyModel(entries=entries, pid=1234)
        gui = self._makeOne(mw)
        response = gui.getFeed()
        self.assertTrue(isinstance(response, Response))
        self.assertEqual(response.content_type, 'application/atom+xml')
        # XXX need more assertions?  Damn trying to test rendered output!

class DummyModel:
    def __init__(self, **kw):
        self.__dict__.update(kw)



import itertools
import os
import time
import threading
try:
    from urllib import quote
except: #pragma NO COVER Py3k
    from urllib.parse import quote

from repoze.debug.ui import is_gui_url
from repoze.debug.ui import DebugGui

class ResponseLoggingMiddleware(object):
    def __init__(self, app, max_bodylen, keep, verbose_logger, trace_logger):
        self.application = app
        self.max_bodylen = max_bodylen
        self.verbose_logger = verbose_logger
        self.trace_logger = trace_logger
        self.keep = keep
        self.entries = []
        self.lock = threading.Lock()
        self.first_request = True
        if hasattr(os, 'getpid'): # pragma: no cover
            self.pid = os.getpid()
        else:
            self.pid = 0 # pragma: no cover

    _now = None
    @property
    def now(self):
        if self._now is None:
            return time.time()
        return self._now

    def __call__(self, environ, start_response):
        if is_gui_url(environ):
            gui = DebugGui(self)
            return gui(environ, start_response)

        request_id = id(environ)
        request_info = self.get_request_info(environ)
        request_info['begin'] = self.now
        self.log_request_begin(request_id, request_info)

        entry = {}
        entry['id'] = request_id
        entry['request'] = request_info

        if self.keep:
            self.lock.acquire()
            try:
                if len(self.entries) >= self.keep:
                    self.entries.pop(0)
                self.entries.append(entry)
            finally:
                self.lock.release()

        catch_response = []
        written = []

        def replace_start_response(status, headers, exc_info=None):
            catch_response.append([status, headers])
            start_response(status, headers, exc_info)
            return written.append

        app_iter = self.application(environ, replace_start_response)
        received_response = self.now

        if catch_response:
            status, headers = catch_response[0]
        else:
            status = '500 Start Response Not Called'
            headers = []

        response_info = self.get_response_info(status, headers)
        response_info['begin'] = received_response

        entry['response'] = response_info

        close = getattr(app_iter, 'close', None)

        body = itertools.chain(written, app_iter)
        body = self.log_response(request_id, request_info, response_info, body,
                                 close)

        return body

    def get_request_info(self, environ):
        info = {}
        supplement = Supplement(self, environ)
        request_data = supplement.extraData()
        method = environ.get('REQUEST_METHOD', 'GET')
        info['method'] = method
        info['url'] = supplement.source_url
        info['cgi_variables'] = []
        info['wsgi_variables'] = []
        info['body'] = ''
        if 'wsgi.input' in environ:
            info['body'] = environ['wsgi.input'].read()
            environ['wsgi.input'].seek(0)
        for k, v in sorted(request_data[('extra', 'CGI Variables')].items()):
            info['cgi_variables'].append((k, v))
        for k, v in sorted(request_data[('extra', 'WSGI Variables')].items()):
            if not 'repoze.debug' in k:
                info['wsgi_variables'].append((k, v))
        return info

    def log_request_begin(self, request_id, info):
        out = []
        begin = info['begin']
        t = time.ctime(begin)
        out.append('--- begin REQUEST for %s at %s ---' % (request_id, t))
        method_and_url = '%s %s' % (info['method'], info['url'])
        out.append('URL: %s' % method_and_url)
        out.append('CGI Variables')
        for k, v in info['cgi_variables']:
            out.append('  %s: %s' % (k, v))
        out.append('WSGI Variables')
        for k, v in info['wsgi_variables']:
            out.append('  %s: %s' % (k, v))
        out.append('Body:')
        out.append(info['body'].decode('latin1'))
        out.append('Bodylen: %s' % len(info['body']))
        out.append('--- end REQUEST for %s ---' % request_id)
        self.verbose_logger and self.verbose_logger.info('\n'.join(out))
        self.lock.acquire()
        try:
            if self.first_request:
                if self.trace_logger is not None:
                    info = 'U %s %s %s' % (self.pid, request_id, begin)
                    self.trace_logger.info(info)
                self.first_request = False
        finally:
            self.lock.release()
        if self.trace_logger is not None:
            info = 'B %s %s %s %s' % (self.pid, request_id, begin,
                                      method_and_url)
            self.trace_logger.info(info)

    def get_response_info(self, status, headers):
        info = {}
        cl = header_value(headers, 'content-length')
        info['headers'] = list(headers)
        try:
            cl = int(cl)
        except (TypeError, ValueError):
            cl = None
        info['content-length'] = cl
        info['status'] = status
        return info

    def log_response(self, request_id, request_info, response_info, body,close):
        out = []
        begin = response_info['begin']
        t = time.ctime(begin)
        status = response_info['status'].split(' ', 1)[0]
        cl = response_info['content-length']
        if self.trace_logger is not None:
            info = 'A %s %s %s %s %s' % (self.pid, request_id, begin,
                                         status, cl)
            self.trace_logger.info(info)
        out.append('--- begin RESPONSE for %s at %s ---' % (request_id, t))
        out.append('URL: %s %s' % (request_info['method'], request_info['url']))
        out.append('Status: %s' % response_info['status'])
        out.append('Response Headers')
        for k, v in response_info['headers']:
            out.append('  %s: %s' % (k, v))
        bodyout = ''
        bodylen = 0
        remaining = None
        for chunk in body:
            if self.max_bodylen:
                remaining = max(0, self.max_bodylen - len(bodyout))
            bodyout += chunk[:remaining]
            bodylen += len(chunk)
            yield chunk
        if bodylen > self.max_bodylen:
            bodyout += ' ... (truncated at %s bytes)' % self.max_bodylen
        out.append('Body:\n' + bodyout)
        out.append('Bodylen: %s' % bodylen)
        if cl is not None:
            if bodylen != cl:
                out.append(
                    'WARNING-1: bodylen (%s) != Content-Length '
                    'header value (%s)' % (bodylen, cl))
        response_info['body'] = bodyout
        end = response_info['end'] = self.now
        duration = response_info['end'] - request_info['begin']
        out.append('--- end RESPONSE for %s (%0.2f seconds) ---' % (
            request_id, duration))
        self.verbose_logger and self.verbose_logger.info('\n'.join(out))
        if self.trace_logger is not None:
            info = 'E %s %s %s %s' % (self.pid, request_id, end, bodylen)
            self.trace_logger.info(info)
        if close is not None:
            close()

class SuffixMultiplier:
    # d is a dictionary of suffixes to integer multipliers.  If no suffixes
    # match, default is the multiplier.  Matches are case insensitive.  Return
    # values are in the fundamental unit.
    def __init__(self, d, default=1):
        self._d = d
        self._default = default
        # all keys must be the same size
        self._keysz = None
        for k in d.keys():
            if self._keysz is None:
                self._keysz = len(k)
            else:
                assert self._keysz == len(k)

    def __call__(self, v):
        v = v.lower()
        for s, m in self._d.items():
            if v[-self._keysz:] == s:
                return int(v[:-self._keysz]) * m
        return int(v) * self._default

byte_size = SuffixMultiplier({'kb': 1024,
                              'mb': 1024*1024,
                              'gb': 1024*1024*1024,})

def make_middleware(app,
                    global_conf,
                    verbose_log=None,
                    trace_log=None,
                    max_bodylen='3KB',
                    max_logsize='100MB',
                    backup_count='10',
                    keep='100',
                    ):
    """ Paste filter-app converter """
    backup_count = int(backup_count)
    max_bytes = byte_size(max_logsize)
    max_bodylen = byte_size(max_bodylen)
    keep = int(keep)
    from logging import Logger
    from logging.handlers import RotatingFileHandler

    if verbose_log:
        handler = RotatingFileHandler(verbose_log, maxBytes=max_bytes,
                                      backupCount=backup_count)
        verbose_log = Logger('repoze.debug.verboselogger')
        verbose_log.handlers = [handler]

    if trace_log:
        handler = RotatingFileHandler(trace_log, maxBytes=max_bytes,
                                      backupCount=backup_count)
        trace_log = Logger('repoze.debug.tracelogger')
        trace_log.handlers = [handler]

    return ResponseLoggingMiddleware(app, max_bodylen, keep, verbose_log,
                                     trace_log)


class Supplement(object):
    """ Display standard WSGI information in the traceback.

    Forked from 'Paste 1.7.5.1 to get Py3k compatibility.
    """
    def __init__(self, middleware, environ):
        self.middleware = middleware
        self.environ = environ
        self.source_url = construct_url(environ)

    def extraData(self):
        data = {}
        cgi_vars = data[('extra', 'CGI Variables')] = {}
        wsgi_vars = data[('extra', 'WSGI Variables')] = {}
        hide_vars = ['paste.config', 'wsgi.errors', 'wsgi.input',
                     'wsgi.multithread', 'wsgi.multiprocess',
                     'wsgi.run_once', 'wsgi.version',
                     'wsgi.url_scheme']
        for name, value in self.environ.items():
            if name.upper() == name:
                if value:
                    cgi_vars[name] = value
            elif name not in hide_vars:
                wsgi_vars[name] = value
        if self.environ['wsgi.version'] != (1, 0):
            wsgi_vars['wsgi.version'] = self.environ['wsgi.version']
        proc_desc = tuple([int(bool(self.environ[key]))
                           for key in ('wsgi.multiprocess',
                                       'wsgi.multithread',
                                       'wsgi.run_once')])
        wsgi_vars['wsgi process'] = self.process_combos[proc_desc]
        wsgi_vars['application'] = self.middleware.application
        p_conf = self.environ.get('paste.config')
        if p_conf is not None:
            data[('extra', 'Configuration')] = dict(p_conf)
        return data

    process_combos = {
        # multiprocess, multithread, run_once
        (0, 0, 0): 'Non-concurrent server',
        (0, 1, 0): 'Multithreaded',
        (1, 0, 0): 'Multiprocess',
        (1, 1, 0): 'Multi process AND threads (?)',
        (0, 0, 1): 'Non-concurrent CGI',
        (0, 1, 1): 'Multithread CGI (?)',
        (1, 0, 1): 'CGI',
        (1, 1, 1): 'Multi thread/process CGI (?)',
        }

def construct_url(environ):
    """Reconstruct the URL from the WSGI environment.

    You may override SCRIPT_NAME, PATH_INFO, and QUERYSTRING with
    the keyword arguments.

    Forked from 'Paste 1.7.5.1 to get Py3k compatibility.
    """
    url = environ['wsgi.url_scheme'] + '://'

    if environ.get('HTTP_HOST'):
        host = environ['HTTP_HOST']
        port = None
        if ':' in host:
            host, port = host.split(':', 1)
            if environ['wsgi.url_scheme'] == 'https':
                if port == '443':
                    port = None
            elif environ['wsgi.url_scheme'] == 'http':
                if port == '80':
                    port = None
        url += host
        if port:
            url += ':%s' % port
    else:
        url += environ['SERVER_NAME']
        if environ['wsgi.url_scheme'] == 'https':
            if environ['SERVER_PORT'] != '443':
                url += ':' + environ['SERVER_PORT']
        else:
            if environ['SERVER_PORT'] != '80':
                url += ':' + environ['SERVER_PORT']

    url += quote(environ.get('SCRIPT_NAME',''))
    url += quote(environ.get('PATH_INFO',''))
    if environ.get('QUERY_STRING'):
        url += '?' + environ['QUERY_STRING']
    return url

def header_value(headers, name):
    """Return the header's value, or None if no such header.

    If a header appears more than once, all the values of the headers
    are joined with ','.   Note that this is consistent /w RFC 2616
    section 4.2 which states:

        It MUST be possible to combine the multiple header fields
        into one "field-name: field-value" pair, without changing
        the semantics of the message, by appending each subsequent
        field-value to the first, each separated by a comma.

    However, note that the original netscape usage of 'Set-Cookie',
    especially in MSIE which contains an 'expires' date will is not
    compatible with this particular concatination method.

    Forked from 'Paste 1.7.5.1 to get Py3k compatibility.
    """
    name = name.lower()
    result = [value for header, value in headers
              if header.lower() == name]
    if result:
        return ','.join(result)
    else:
        return None



import pdb
import sys

# stolen partly from z3c.evalexception
def PostMortemDebug(application, *ignore_exc):
    """Middleware that catches exceptions and invokes pdb's
    post-mortem debugging facility."""
    def middleware(environ, start_response):
        try:
            return application(environ, start_response)
        except ignore_exc:
            raise
        except:
            pdb.post_mortem(sys.exc_info()[2])
            raise

    return middleware

def make_middleware(app, global_conf, ignore_http_exceptions=True):
    if ignore_http_exceptions:
        try:
            from paste.httpexceptions import HTTPException
        except ImportError: #pragma NO COVER Py3k
            pass
        else: #pragma NO COVER Py3k
            return PostMortemDebug(app, HTTPException)
    return PostMortemDebug(app)
    




##############################################################################
#
# Copyright (c) 2007 Agendaless Consulting and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the BSD-like license at
# http://www.repoze.org/LICENSE.txt.  A copy of the license should accompany
# this distribution.  THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL
# EXPRESS OR IMPLIED WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO,
# THE IMPLIED WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND
# FITNESS FOR A PARTICULAR PURPOSE
#
##############################################################################

__version__ = '1.0'

import os

from setuptools import setup, find_packages

requires = ['WebOb']


here = os.path.abspath(os.path.dirname(__file__))
README = open(os.path.join(here, 'README.rst')).read()
CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()

testing_extras = ['nose', 'coverage']

docs_extras = ['Sphinx']

setup(name='repoze.debug',
      version=__version__,
      description='Forensic debugging WSGI middleware',
      long_description=README + '\n\n' + CHANGES,
      classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Internet :: WWW/HTTP :: Dynamic Content",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Middleware",
        ],
      keywords='wsgi request response debug middleware',
      author="Agendaless Consulting",
      author_email="repoze-dev@lists.repoze.org",
      url="http://www.repoze.org",
      license="BSD-derived (http://www.repoze.org/LICENSE.txt)",
      packages=find_packages(),
      include_package_data=True,
      namespace_packages=['repoze'],
      zip_safe=False,
      tests_require = requires,
      install_requires = requires,
      test_suite="repoze.debug.tests",
      entry_points = """\
        [paste.filter_app_factory]
        responselogger = repoze.debug.responselogger:make_middleware
        canary = repoze.debug.canary:make_middleware
        pdbpm = repoze.debug.pdbpm:make_middleware
        threads = repoze.debug.threads:make_middleware
        [console_scripts]
        wsgirequestprofiler = repoze.debug.scripts.requestprofiler:main
      """,
      extras_require = {
        'testing':  requires + testing_extras,
        'docs':  requires + docs_extras,
      }
)




"""GUI for presenting ways to look at the repoze.debug data

"""
try:
    from html import escape
except ImportError: #pragma NO COVER Pyton3
    from cgi import escape
import mimetypes
import os
import pprint
import time

from webob import Response

_HERE = os.path.abspath(os.path.dirname(__file__))
gui_flag = '__repoze.debug'

def is_gui_url(environ):
    return gui_flag in environ.get('PATH_INFO', '')

def get_mimetype(filename):
    type, encoding = mimetypes.guess_type(filename)
    if type is None and filename.endswith(".xul"):
        return 'application/vnd.mozilla.xul+xml' #pragma NO COVER
    return type or 'application/octet-stream'

class DebugGui(object):

    def __init__(self, middleware):
        self.middleware = middleware
        self.static_dir = os.path.join(_HERE, 'static')

    def __call__(self, environ, start_response):
        """Pick apart this debug URL and return the correct response"""
        path = environ['PATH_INFO']

        if '/static/' in path:
            resp = self.getStatic(path)
        elif gui_flag + '/feed.xml' in path:
            resp = self.getFeed()
        else:
            raise ValueError('No such handler for debug ui: %s', path)

        return resp(environ, start_response)

    def getStatic(self, path):
        fn = path.split('/')[-1]

        if not (fn in os.listdir(self.static_dir)):
            raise ValueError('No such static file %s' % fn)
        
        filename = os.path.join(self.static_dir, fn)
        res = Response(content_type=get_mimetype(filename))
        with open(filename, 'rb') as f:
            res.body = f.read()
        return res

    def _generateFeedTagURI(self, when, pid):
        """ See http//www.taguri.org """
        date = time.strftime('%Y-%m-%d', time.localtime(when))
        pid = self.middleware.pid
        return 'tag:repoze.org,%s:%s' % (date, pid)

    def _generateEntryTagURI(self, entry):
        """ See http//www.taguri.org """
        date = time.strftime('%Y-%m-%d', time.localtime(
            entry['request']['begin']))
        pid = self.middleware.pid
        return 'tag:repoze.org,%s:%s-%s' % (date, entry['id'], pid)

    def getFeed(self):
        """Get XML representing information in the middleware"""

        entries_xml = []

        for entry in self.middleware.entries:
            request = entry['request']
            response = entry.get('response')
            begin = time.localtime(request['begin'])
            entry_id = self._generateEntryTagURI(entry)
            entry_title = '%s %s ' % (request['method'], request['url'])

            short_url = request['url']
            max_url_len = 40
            if len(short_url) > max_url_len:
                prefix = short_url[:9]
                suffix = short_url[-max_url_len+9:]
                short_url = prefix + '...' + suffix
            entry_title = '%s %s ' % (request['method'], short_url)

            # Make the <rz:cgi_variable> nodes into a string
            cgivars = ""
            for k,v in request['cgi_variables']:
                newv = escape(str(v))
                s = cgi_variable_fmt % (k, newv)
                cgivars = cgivars + s

            # Make the <rz:cgi_variable> nodes into a string
            wsgivars = ""
            for k,v in request['wsgi_variables']:
                newv = escape(str(v))
                s = wsgi_variable_fmt % (k, newv)
                wsgivars = wsgivars + s

            # Make the <rz:request> node
            rzrequest = rzrequest_fmt % {
                'begin': request['begin'],
                'cgi_variables': cgivars,
                'wsgi_variables': wsgivars,
                'method': request['method'],
                'url': request['url'],
                'body': escape(request['body']),
                }

            if response is not None:
                # Make the <rz:request> node
                headers = ''
                for k,v in response['headers']:
                    newv = escape(str(v))
                    s = header_fmt % (k, newv)
                    headers = headers + s

                rzresponse = rzresponse_fmt % {
                    'begin': response['begin'],
                    'end': response['end'],
                    'content-length': response['content-length'],
                    'headers': headers,
                    'status': response['status'],
                    'body': escape(response['body']),
                    }
            else:
                rzresponse = ''


            # Make the atom:entry/atom:content node
            content = contentfmt % {
                'logentry_id': entry_id,
                'rzrequest': rzrequest,
                'rzresponse': rzresponse,
                }

            entry_xml = entryfmt % {
                'entry_id':entry_id,
                'entry_title':escape(entry_title),
                'updated':time.strftime('%Y-%m-%dT%H:%M:%SZ', begin),
                'summary':escape(pprint.pformat(entry)),
                'content':content,
                }
            entries_xml.append(entry_xml)

        now = time.time()

        body = feedfmt % {
            'title':'repoze.debug feed for pid %s' % self.middleware.pid,
            'entries':'\n'.join(entries_xml),
            'feed_id':self._generateFeedTagURI(now, self.middleware.pid),
            'updated':time.strftime('%Y-%m-%dT%H:%M:%SZ', time.localtime(now)),
            }

        resp = Response(content_type='application/atom+xml', body=body)
        return resp

feedfmt = """\
<?xml version="1.0" encoding="utf-8"?>
<atom:feed xmlns:atom="http://www.w3.org/2005/Atom">

  <atom:title>%(title)s</atom:title>
  <atom:link href="http://example.org/"/>
  <atom:updated>%(updated)s</atom:updated>
  <atom:author>
    <atom:name></atom:name>
  </atom:author>
  <atom:id>%(feed_id)s</atom:id>
  %(entries)s
</atom:feed>
"""

entryfmt = """\
  <atom:entry>
    <atom:id>%(entry_id)s</atom:id>
    <atom:title>%(entry_title)s</atom:title>
    <atom:link href="http://example.org/2003/12/13/atom03"/>
    <atom:updated>%(updated)s</atom:updated>
    <atom:summary>%(summary)s</atom:summary>
    <atom:content>%(content)s</atom:content>
  </atom:entry>
"""

contentfmt = """\
<rz:logentry id="%(logentry_id)s" xmlns:rz="http://repoze.org/namespace">
  %(rzrequest)s
  %(rzresponse)s
</rz:logentry>
"""

cgi_variable_fmt = """<rz:cgi_variable name="%s">%s</rz:cgi_variable>
"""

wsgi_variable_fmt = """<rz:wsgi_variable name="%s">%s</rz:wsgi_variable>
"""

rzrequest_fmt = """<rz:request>
  <rz:begin>%(begin)s</rz:begin>
  %(cgi_variables)s
  <rz:method>%(method)s</rz:method>
  <rz:url>%(url)s</rz:url>
  %(wsgi_variables)s  
  <rz:body>
%(body)s
  </rz:body>
</rz:request>
"""

header_fmt = """<rz:header name="%s">%s</rz:header>
"""

rzresponse_fmt = """<rz:response>
  <rz:begin>%(begin)s</rz:begin>
  <rz:end>%(end)s</rz:end>
  <rz:content-length>%(content-length)s</rz:content-length>
  %(headers)s
  <rz:status>%(status)s</rz:status>
  <rz:body>
%(body)s
  </rz:body>
</rz:response>
"""



##############################################################################
#
# Copyright (c) 2002 Zope Corporation and Contributors. All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE
#
##############################################################################
"""Trace log profiler script

$Id: requestprofiler.py 40218 2005-11-18 14:39:19Z andreasjung $
"""
import cPickle
import getopt
try:
    import gzip
except:
    gzip = None
import sys
import time

class ProfileException(Exception):
    pass

class Request:
    def __init__(self):
        self.url = None
        self.start = None
        self.method = None
        self.t_recdinput = None
        self.isize = 0
        self.t_recdoutput = None
        self.osize = None
        self.httpcode = None
        self.t_end = None
        self.elapsed = None
        self.active = 0

    def put(self, code, t, desc):
        if code not in ('A', 'B', 'I', 'E'):
            raise "unknown request code %s" % code
        if code == 'B':
            self.start = t
            self.method, self.url = desc.strip().split()
        elif code == "A":
            self.t_recdinput = t
            self.t_recdoutput = t
            self.httpcode, self.osize = desc.strip().split()
        elif code == 'E':
            self.t_end = t
            self.elapsed = self.t_end - self.start

    def isfinished(self):
        return not self.elapsed is None

    def prettystart(self):
        if self.start is not None:
            t = time.localtime(self.start)
            return time.strftime('%Y-%m-%dT%H:%M:%S', t)
        else:
            return "NA"

    def shortprettystart(self):
        if self.start is not None:
            t = time.localtime(self.start)
            return time.strftime('%H:%M:%S', t)
        else:
            return -1

    def win(self):
        if self.t_recdinput is not None and self.start is not None:
            return self.t_recdinput - self.start
        else:
            return -1

    def wout(self):
        if self.t_recdoutput is not None and self.t_recdinput is not None:
            return self.t_recdoutput - self.t_recdinput
        else:
            return -1

    def wend(self):
        if self.t_end is not None and self.t_recdoutput is not None:
            return self.t_end - self.t_recdoutput
        else:
            return -1

    def endstage(self):
        if self.t_end is not None:
            stage = "E"
        elif self.t_recdoutput is not None:
            stage = "A"
        elif self.t_recdinput is not None:
            stage = "I"
        else:
            stage = "B"
        return stage

    def total(self):
        stage = self.endstage()
        if stage == "B":
            return 0
        if stage == "I":
            return self.t_recdinput - self.start
        if stage == "A":
            return self.t_recdoutput - self.start
        if stage == "E":
            return self.elapsed

    def prettyisize(self):
        if self.isize is not None:
            return self.isize
        else:
            return -1

    def prettyosize(self):
        if self.osize is not None:
            return self.osize
        else:
            return -1

    def prettyhttpcode(self):
        if self.httpcode is not None:
            return self.httpcode
        else:
            return "NA"

    def __str__(self):
        fmt = "%19s %5.2f %5.2f %5.2f %5.2f %1s %7s %4s %4s %s"
        body = (
            self.prettystart(), self.win(), self.wout(), self.wend(),
            self.total(), self.endstage(), self.prettyosize(),
            self.prettyhttpcode(), self.active, self.url
            )
        return fmt % body

    def getheader(self):
        fmt = "%19s %5s %5s %5s %5s %1s %7s %4s %4s %s"
        body = ('Start', 'WIn', 'WOut', 'WEnd', 'Tot', 'S', 'OSize',
                'Code', 'Act', 'URL')
        return fmt % body

class StartupRequest(Request):
    def endstage(self):
        return "U"

    def total(self):
        return 0

class Cumulative:
    def __init__(self, url):
        self.url = url
        self.times = []
        self.hangs = 0
        self.allelapsed = None

    def put(self, request):
        elapsed = request.elapsed
        if elapsed is None:
            self.hangs = self.hangs + 1
        else:
            self.times.append(elapsed)

    def all(self):
        if self.allelapsed is None:
            self.allelapsed = []
            for elapsed in self.times:
                self.allelapsed.append(elapsed)
            self.allelapsed.sort()
        return self.allelapsed

    def __str__(self):
        fmt = "%5s %5s %8.2f %5.2f %5.2f %5.2f %5.2f %s"
        body = (
            self.hangs, self.hits(), self.total(), self.max(), self.min(),
            self.median(), self.mean(), self.url
            )
        return fmt % body

    def getheader(self):
        fmt = '%5s %5s %8s %5s %5s %5s %5s %s'
        return fmt % ('Hangs', 'Hits', 'Total', 'Max', 'Min', 'Med',
                      'Mean', 'URL')

    def hits(self):
        return len(self.times)

    def max(self):
        all = self.all()
        if not all:
            return 0
        return max(all)

    def min(self):
        all = self.all()
        if not all:
            return 0
        return min(all)

    def mean(self):
        l = len(self.times)
        if l == 0:
            return 0
        else:
            t = self.total()
            return t/l

    def median(self):
        all = self.all()
        l = len(all)
        if l == 0:
            return 0
        else:
            if l == 1:
                return all[0]
            elif l % 2 != 0:
                i = l/2 + 1
                return all[i]
            else:
                i = l/2 - 1
                i2 = i + 1
                v1 = all[i]
                v2 = all[i2]
                return (v1 + v2) / 2

    def total(self):
        t = 0
        all = self.all()
        for elapsed in all:
            if elapsed is None:
                continue
            t = t + elapsed
        return float(t)

def parselogline(line):
    tup = line.split(None, 4)
    if len(tup) == 4:
        code, pid, id, timestr = tup
        return code, pid, id, timestr, ''
    elif len(tup) == 5:
        return tup
    else:
        return None

def get_earliest_file_data(files):
    temp = {}
    earliest_fromepoch = 0
    earliest = None
    retn = None
    for file in files:
        line = file.readline()
        if not line:
            continue
        linelen = len(line)
        line = line.strip()
        tup = parselogline(line)
        if tup is None:
            print "Could not interpret line: %s" % line
            continue
        code, pid, id, timestr, desc = tup
        timestr = timestr.strip()
        fromepoch = float(timestr)
        temp[file] = linelen
        if earliest_fromepoch == 0 or fromepoch < earliest_fromepoch:
            earliest_fromepoch = fromepoch
            earliest = file
            retn = [code, pid, id, fromepoch, desc]

    for file, linelen in temp.items():
        if file is not earliest:
            file.seek(file.tell() - linelen)

    return retn

def get_requests(files, start=None, end=None, statsfname=None,
                 writestats=None, readstats=None):
    finished = []
    unfinished = {}
    if readstats:
        fp = open(statsfname, 'r')
        u = cPickle.Unpickler(fp)
        requests = u.load()
        fp.close()
        del u
        del fp
    else:
        while 1:
            tup = get_earliest_file_data(files)
            if tup is None:
                break
            code, pid, id, fromepoch, desc = tup
            if start is not None and fromepoch < start:
                continue
            if end is not None and fromepoch > end:
                break
            if code == 'U':
                # restart
                for upid, uid in list(unfinished.keys()):
                    if upid == pid:
                        val = unfinished[(upid, uid)]
                        finished.append(val)
                        del unfinished[(upid, uid)]
                request = StartupRequest()
                request.url = desc
                request.start = fromepoch
                finished.append(request)
                continue
            request = unfinished.get((pid, id))
            if request is None:
                if code != "B":
                    continue # garbage at beginning of file
                request = Request()
                for pending_req in unfinished.values():
                    pending_req.active = pending_req.active + 1
                unfinished[(pid, id)] = request
            try:
                request.put(code, fromepoch, desc)
            except:
                print "Unable to handle entry: %s %s %s"%(code, fromepoch, desc)
            if request.isfinished():
                del unfinished[(pid, id)]
                finished.append(request)

        finished.extend(unfinished.values())
        requests = finished

        if writestats:
            fp = open(statsfname, 'w')
            p = cPickle.Pickler(fp)
            p.dump(requests)
            fp.close()
            del p
            del fp

    return requests

def analyze(requests, top, sortf, start=None, end=None, mode='cumulative',
            resolution=60, urlfocusurl=None, urlfocustime=60, verbose=False):

    if mode == 'cumulative':
        cumulative = {}
        for request in requests:
            if not isinstance(request, StartupRequest):
                url = request.url
                stats = cumulative.get(url)
                if stats is None:
                    stats = Cumulative(url)
                    cumulative[url] = stats
                stats.put(request)
        requests = cumulative.values()
        requests.sort(sortf)
        write(requests, top, verbose)

    elif mode=='timed':
        computed_start = requests[0].start
        computed_end = requests[-1].t_end
        if start and end:
            timewrite(requests,start,end,resolution)
        if start and not end:
            timewrite(requests,start,computed_end,resolution)
        if end and not start:
            timewrite(requests,computed_start,end,resolution)
        if not end and not start:
            timewrite(requests,computed_start,computed_end,resolution)

    elif mode == 'urlfocus':
        requests.sort(sortf)
        urlfocuswrite(requests, urlfocusurl, urlfocustime)

    else:
        requests.sort(sortf)
        write(requests, top, verbose)

def urlfocuswrite(requests, url, t):
    l = []
    i = 0
    for request in requests:
        if request.url == url: l.append(i)
        i = i + 1
    before = {}
    after = {}
    x = 0
    for n in l:
        x = x + 1
        r = requests[n]
        start = r.start
        earliest = start - t
        latest = start + t
        print 'URLs invoked %s seconds before and after %s (#%s, %s)' % \
              (t, url, x, r.shortprettystart())
        print '---'
        i = -1
        for request in requests:
            i = i + 1
            if request.start < earliest: continue
            if request.start > latest: break
            if n == i: # current request
                print '%3d' % (request.start - start),
                print '%s' % (request.shortprettystart()),
                print request.url
                continue
            if request.start <= start:
                if before.get(i):
                    before[i] = before[i] + 1
                else:
                    before[i] = 1
            if request.start > start:
                if after.get(i):
                    after[i] = after[i] + 1
                else:
                    after[i] = 1
            print '%3d' % (request.start - start),
            print '%s' % (request.shortprettystart()),
            print request.url
        print
    print ('Summary of URLs invoked before (and at the same time as) %s '
           '(times, url)' % url)
    before = before.items()
    before.sort()
    for k,v in before:
        print v, requests[k].url
    print
    print 'Summary of URLs invoked after %s (times, url)' % url
    after = after.items()
    after.sort()
    for k,v in after:
        print v, requests[k].url

def write(requests, top=0, verbose=False):
    if len(requests) == 0:
        print "No data."
        return
    i = 0
    header = requests[0].getheader()
    print header
    for stat in requests:
        i = i + 1
        if verbose:
            print str(stat)
        else:
            print str(stat)[:78]
        if i == top:
            break

def getdate(val):
    try:
        val = val.strip()
        year, month, day = int(val[:4]), int(val[5:7]), int(val[8:10])
        hour,minute,second=int(val[11:13]),int(val[14:16]),int(val[17:19])
        t = time.mktime((year, month, day, hour, minute, second, 0, 0, -1))
        return t
    except:
        raise ProfileException, "bad date %s" % val

def timewrite(requests, start, end, resolution):
    print "Start: %s    End: %s   Resolution: %d secs" % \
        (tick2str(start), tick2str(end), resolution)
    print "-" * 78
    print
    print "Date/Time                #requests requests/second"

    d = {}
    max = 0
    min = None
    for r in requests:
        t = r.start
        slice = t - (t % resolution)
        if slice > max:
            max = slice
        if (min is None) or (slice < min):
            min = slice
        if d.has_key(slice):
            d[slice] = d[slice] + 1
        else:
            d[slice] = 1

    num = 0
    hits = 0
    avg_requests = None
    max_requests = 0
    for slice in range(int(min), int(max), resolution):
        num = d.get(slice, 0)
        if num>max_requests:
            max_requests = num
        hits = hits + num

        if avg_requests is None:
            avg_requests = num
        else:
            avg_requests = (avg_requests + num) / 2

        s = tick2str(slice)
        s = s + "     %6d         %4.2lf" % (num,num*1.0/resolution)
        print s

    print '='*78
    print " Peak:                  %6d         %4.2lf" % \
        (max_requests,max_requests*1.0/resolution)
    print "  Avg:                  %6d         %4.2lf" % \
        (avg_requests,avg_requests*1.0/resolution)
    print "Total:                  %6d          n/a " % (hits)

def tick2str(t):
    return time.strftime('%Y-%m-%dT%H:%M:%S', time.localtime(t))

def codesort(v1, v2):
    v1 = v1.endstage()
    v2 = v2.endstage()
    if v1 == v2:
        return 0

    if v1 == "B":
        return -1 # v1 is smaller than v2
    if v1 == "I":
        if v2 == "B": return 1 # v1 is larger than v2
        else: return -1
    if v1 == "A":
        if v2 in ['B', 'I']: return 1
        else: return -1
    if v1 == "E":
        return 1

class Sort:
    def __init__(self, fname, ascending=0):
        self.fname = fname
        self.ascending = ascending

    def __call__(self, i1, i2):
        f1 = getattr(i1, self.fname)
        f2 = getattr(i2, self.fname)
        if callable(f1): f1 = f1()
        if callable(f2): f2 = f2()
        if f1 < f2:
            if self.ascending: return -1
            else: return 1
        elif f1 == f2:
            return 0
        else:
            if self.ascending: return 1
            else: return -1

def detailedusage():
    details = usage(0)
    pname = sys.argv[0]
    details = details + """
Reports are of four types: cumulative, detailed, timed, or urlfocus.
The default is cumulative. Data is taken from one or more repoze.debug
trace logs or from a preprocessed statistics file.

For cumulative reports, each line in the profile indicates information
about a URL collected via a detailed request log.

For detailed reports, each line in the profile indicates information about
a single request.

For timed reports, each line in the profile indicates informations about
the number of requests and the number of requests/second for a period of time.

For urlfocus reports, ad-hoc information about requests surrounding the
specified url is given.

Each 'filename' is a path to a trace log that contains detailed
request data.  Multiple input files can be analyzed at the same time
by providing the path to each file.  (Analyzing multiple trace log
files at once is useful if you have more than one machine running your
application and you'd like to get an overview of all logs on those
machines).

If you wish to make multiple analysis runs against the same input
data, you may want to use the --writestats option.  The --writestats
option creates a file which holds preprocessed data representing the
specfified input files.  Subsequent runs (for example with a different
sort spec) will be much faster if the --readstats option is used to
specify a preprocessed stats file instead of actual input files
because the logfile parse step is skipped.

If a 'sort' value is specified, sort the profile info by the spec.
The sort order is descending unless indicated.  The default cumulative
sort spec is 'total'.  The default detailed sort spec is 'start'.

For cumulative reports, the following sort specs are accepted:

  'hits'        -- the number of hits against the method
  'hangs'       -- the number of unfinished requests to the method
  'max'         -- the maximum time in secs taken by a request to this method
  'min'         -- the minimum time in secs taken by a request to this method
  'mean'        -- the mean time in secs taken by a request to this method
  'median'      -- the median time in secs taken by a request to this method
  'total'       -- the total time in secs across all requests to this method
  'url'         -- the URL/method name (ascending)

For detailed (non-cumulative) reports, the following sort specs are accepted:

  'start'       -- the start time of the request to repoze.debug (ascending)
  'win'         -- the num of secs repoze.debug spent waiting for input
  'wout'        -- the secs repoze.debug spent waiting for output from app
  'wend'        -- the secs repoze.debug spent sending data to server
  'total'       -- the secs taken for the request from begin to end
  'endstage'    -- the last successfully completed request stage (B, I, A, E)
  'osize'       -- the size in bytes of output provided by repoze.debug
  'httpcode'    -- the HTTP response code provided by the app (ascending)
  'active'      -- total num of requests pending at the end of this request
  'url'         -- the URL  (ascending)

For timed and urlfocus reports, there are no sort specs allowed.

If the 'top' argument is specified, only report on the top 'n' entries
in the profile (as per the sort). The default is to show all data in
the profile.

If the 'verbose' argument is specified, do not trim url to fit into 80
cols.

If the 'today' argument is specified, limit results to hits received
today.

If the 'daysago' argument is specified, limit results to hits received
n days ago.

The 'resolution' argument is used only for timed reports and specifies
the number of seconds between consecutive lines in the report (default
is 60 seconds).

The 'urlfocustime' argument is used only for urlfocus reports and
specifies the number of seconds to target before and after the URL
provided in urlfocus mode.  (default is 10 seconds).

If the 'start' argument is specified in the form 'DD/MM/YYYY HH:MM:SS'
(local time), limit results to hits received after this date/time.

If the 'end' argument is specified in the form 'DD/MM/YYYY HH:MM:SS'
(local time), limit results to hits received before this date/time.

'start' and 'end' arguments are not honored when request stats are obtained
via the --readstats argument.

Examples:

  %(pname)s debug.log

    Show cumulative report statistics for information in the file 'debug.log',
    by default sorted by 'total'.

  %(pname)s debug.log --detailed

    Show detailed report statistics sorted by 'start' (by default).

  %(pname)s debug.log debug2.log --detailed

    Show detailed report statistics for both logs sorted by 'start'
    (by default).

  %(pname)s debug.log --cumulative --sort=mean --today --verbose

    Show cumulative report statistics sorted by mean for entries in the log
    which happened today, and do not trim the URL in the resulting report.

  %(pname)s debug.log --cumulative --sort=mean --daysago=3 --verbose

    Show cumulative report statistics sorted by mean for entries in the log
    which happened three days ago, and do not trim the URL in the resulting report.

  %(pname)s debug.log --urlfocus='/manage_main' --urlfocustime=60

    Show 'urlfocus' report which displays statistics about requests
    surrounding the invocation of '/manage_main'.  Focus on the time periods
    60 seconds before and after each invocation of the '/manage_main' URL.

  %(pname)s debug.log --detailed --start='2001/05/10 06:00:00'
    --end='2001/05/11 23:00:00'

    Show detailed report statistics for entries in 'debug.log' which
    begin after 6am local time on May 10, 2001 and which end before
    11pm local time on May 11, 2001.

  %(pname)s debug.log --timed --resolution=300 --start='2001/05/10 06:00:00'
    --end='2001/05/11 23:00:00'

    Show timed report statistics for entries in the log for one day
    with a resolution of 5 minutes

  %(pname)s debug.log --top=100 --sort=max

    Show cumulative report of the the 'top' 100 methods sorted by maximum
    elapsed time.

  %(pname)s debug.log debug2.log --writestats='requests.stat'

    Write stats file for debug.log and debug2.log into 'requests.stat' and
    show default report.

  %(pname)s --readstats='requests.stat' --detailed

    Read from 'requests.stat' stats file (instead of actual -M log files)
    and show detailed report against this data.""" % {'pname':pname}
    return details

def usage(basic=1):
    usage = (
        """
Usage: %s filename1 [filename2 ...]
          [--cumulative | --detailed | [--timed --resolution=seconds]]
          [--sort=spec]
          [--top=n]
          [--verbose]
          [--today | [--start=date] [--end=date] | --daysago=n ]
          [--writestats=filename | --readstats=filename]
          [--urlfocus=url]
          [--urlfocustime=seconds]
          [--help]

Provides a profile of one or more repoze.debug trace log files.
""" % sys.argv[0]
        )
    if basic == 1:
        usage = usage + """
If the --help argument is given, detailed usage docs are provided."""
    return usage

def main():
    if len(sys.argv) == 1:
        print usage()
        sys.exit(0)
    if sys.argv[1] == '--help':
        print detailedusage()
        sys.exit(0)
    mode = 'cumulative'
    sortby = None
    trim = 0
    top = 0
    verbose = 0
    start = None
    end = None
    resolution=60
    urlfocustime=10
    urlfocusurl=None
    statsfname = None
    readstats = 0
    writestats = 0

    files = []
    i = 1
    for arg in sys.argv[1:]:
        if arg[:2] != '--':
            if arg[-3:] == '.gz':
                if gzip is None:
                    raise ValueError('No gzip support to ungzip %s' % arg)
                files.append(gzip.GzipFile(arg,'r'))
            else:
                files.append(open(arg))
            sys.argv.remove(arg)
            i = i + 1

    try:
        opts, extra = getopt.getopt(
            sys.argv[1:], '', ['sort=', 'top=', 'help', 'verbose', 'today',
                               'cumulative', 'detailed', 'timed','start=',
                               'end=','resolution=', 'writestats=','daysago=',
                               'readstats=','urlfocus=','urlfocustime=']
            )
        for opt, val in opts:

            if opt=='--readstats':
                statsfname = val
                readstats = 1
            elif opt=='--writestats':
                statsfname = val
                writestats = 1
            if opt=='--sort':
                sortby = val
            if opt=='--top':
                top=int(val)
            if opt=='--help':
                print detailedusage()
                sys.exit(0)
            if opt=='--verbose':
                verbose = 1
            if opt=='--resolution':
                resolution=int(val)
            if opt=='--today':
                now = time.localtime(time.time())
                # for testing - now = (2001, 04, 19, 0, 0, 0, 0, 0, -1)
                start = list(now)
                start[3] = start[4] = start[5] = 0
                start = time.mktime(start)
                end = list(now)
                end[3] = 23; end[4] = 59; end[5] = 59
                end = time.mktime(end)
            if opt=='--daysago':
                now = time.localtime(time.time() - int(val)*3600*24 )
                # for testing - now = (2001, 04, 19, 0, 0, 0, 0, 0, -1)
                start = list(now)
                start[3] = start[4] = start[5] = 0
                start = time.mktime(start)
                end = list(now)
                end[3] = 23; end[4] = 59; end[5] = 59
                end = time.mktime(end)
            if opt=='--start':
                start = getdate(val)
            if opt=='--end':
                end = getdate(val)
            if opt=='--detailed':
                mode='detailed'
                d_sortby = sortby
            if opt=='--cumulative':
                mode='cumulative'
            if opt=='--timed':
                mode='timed'
            if opt=='--urlfocus':
                mode='urlfocus'
                urlfocusurl = val
            if opt=='--urlfocustime':
                urlfocustime=int(val)

        validcumsorts = ['url', 'hits', 'hangs', 'max', 'min', 'median',
                         'mean', 'total']
        validdetsorts = ['start', 'win', 'wout', 'wend', 'total',
                         'endstage', 'isize', 'osize', 'httpcode',
                         'active', 'url']

        if mode == 'cumulative':
            if sortby is None: sortby = 'total'
            assert sortby in validcumsorts, (sortby, mode, validcumsorts)
            if sortby in ['url']:
                sortf = Sort(sortby, ascending=1)
            else:
                sortf = Sort(sortby)
        elif mode == 'detailed':
            if sortby is None: sortby = 'start'
            assert sortby in validdetsorts, (sortby, mode, validdetsorts)
            if sortby in ['start', 'url', 'httpcode']:
                sortf = Sort(sortby, ascending=1)
            elif sortby == 'endstage':
                sortf = codesort
            else:
                sortf = Sort(sortby)
        elif mode=='timed':
            sortf = None
        elif mode=='urlfocus':
            sortf = Sort('start', ascending=1)
        else:
            raise 'Invalid mode'

        req=get_requests(files, start, end, statsfname, writestats, readstats)
        analyze(req, top, sortf, start, end, mode, resolution, urlfocusurl,
                urlfocustime, verbose)

    except AssertionError as val:
        a = "%s is not a valid %s sort spec, use one of %s"
        print a % (val[0], val[1], val[2])
        sys.exit(0)
    except getopt.error as val:
        print val
        sys.exit(0)
    except ProfileException as val:
        print val
        sys.exit(0)
    except SystemExit:
        sys.exit(0)
    except:
        import traceback
        traceback.print_exc()
        print usage()
        sys.exit(0)

if __name__ == '__main__':
    main()

